/*!
 * \file
 * \brief Builds the Hockney, PLogP and LMO models and estimates the execution time of point-to-point and
 * collective communication operations.
 * Arguments are described in the MPIBlib manual.
 * 
 * \b model.plot draws the graph observation vs predictions.
 * - input: model.out (model output for different models), p2p.out (p2p output)
 * - output: p2p.eps (0-1)
 */

#include "cpm.h"
#include <string.h>
#include <getopt.h>
#include <gsl/gsl_statistics_double.h>

int main(int argc, char** argv) {
	MPI_Init(&argc, &argv);
	MPI_Comm comm = MPI_COMM_WORLD;
#ifndef MPIB_DEBUG
	MPIB_Comm(MPI_COMM_WORLD, &comm);
	if (comm == MPI_COMM_NULL) {
		MPI_Finalize();
		return 0;
	}
#endif

	int exit;
	MPIB_getopt_help_default(&exit);
	MPIB_msgset msgset;
	MPIB_getopt_msgset_default(&msgset);
	MPIB_precision precision;
	MPIB_getopt_precision_default(&precision);
	int parallel;
	MPIB_getopt_p2p_default(&parallel);
	char model_name[256] = "";
	void* model = NULL;
	FILE* istream = NULL;
	FILE* ostream = stdout;

	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (rank == 0) {
		int c;
		char options[256] = "C:i:o:";
		strcat(options, MPIB_getopt_help_options());
		strcat(options, MPIB_getopt_msgset_options());
		strcat(options, MPIB_getopt_precision_options());
		strcat(options, MPIB_getopt_p2p_options());
		while ((c = getopt(argc, argv, options)) != -1) {
			MPIB_getopt_help_optarg(c, "model", &exit);
			MPIB_getopt_precision_optarg(c, &precision);
			MPIB_getopt_msgset_optarg(c, &msgset);
			MPIB_getopt_p2p_optarg(c, &parallel);
			switch (c) {
				case 'h':
					fprintf(
						stderr,
"-C S			communication model (required): Hockney%s%s\n\
-i S			input file (optional)\n\
-o S			output file (default: stdout)\n",
						", PLogP",
#ifdef HAVE_LIBR
						", LMO"
#else
						""
#endif /*HAVE_LIBR*/
					);
					break;
				case 'C':
					strcpy(model_name, optarg);
					break;
				case 'i':
					istream = fopen(optarg, "r");
					if (istream == NULL)
						fprintf(stderr, "Cannot open file %s for reading\n", optarg);
					break;
				case 'o':
					ostream = fopen(optarg, "w");
					if (ostream == NULL)
						fprintf(stderr, "Cannot open file %s for writing\n", optarg);
					break;
			}
		}
	}
	MPIB_getopt_help_bcast(&exit, 0, comm);
	if (exit) {
		MPI_Finalize();
		return 0;
	}
	
	MPIB_getopt_msgset_bcast(&msgset, 0, comm);
	MPIB_getopt_precision_bcast(&precision, 0, comm);
	MPIB_getopt_p2p_bcast(&parallel, 0, comm);

	int len;
	len = strlen(model_name) + 1;
	MPI_Bcast(&len, 1, MPI_INT, 0, comm);
	MPI_Bcast(&model_name, len, MPI_CHAR, 0, comm);
	int build = istream == NULL;
	MPI_Bcast(&build, 1, MPI_INT, 0, comm);
	if (build)
	{
		double time;
		if (!strcmp(model_name, "Hockney")) {
			time = MPI_Wtime();
			Hockney_estimate_regression(comm, msgset, precision, parallel, (Hockney_model**)&model);
			//Hockney_estimate(comm, msgset.max_size, precision, parallel, (Hockney_model**)&model);
			time = MPI_Wtime() - time;
		}
		if (!strcmp(model_name, "LinInterp")) {
			time = MPI_Wtime();
			LinInterp_estimate(comm, precision,msgset, parallel, (LinInterp_model**)&model);
			time = MPI_Wtime() - time;
		}
#ifdef HAVE_LIBR
		if (!strcmp(model_name, "LMO")) {
			if (rank == 0) {
				exit = LMO_initR();
				if (exit)
					fprintf(stderr, "R cannot be initialized\n");
			}
			MPI_Bcast(&exit, 1, MPI_INT, 0, comm);
			if (exit) {
				MPI_Finalize();
				return 0;
			}

			time = MPI_Wtime();
			LMO_estimate(comm, precision, msgset, parallel, (LMO_model**)&model);
			time = MPI_Wtime() - time;

			if (rank == 0)
				LMO_endR();
		}
#endif /*HAVE_LIBR*/
		if (!strcmp(model_name, "PLogP")) {
			time = MPI_Wtime();
			PLogP_estimate(comm, precision, msgset, parallel, (PLogP_model**)&model); 
			time = MPI_Wtime() - time;
		}
		if (rank == 0) {
			exit = model == NULL;
			if (exit)
				fprintf(stderr, "Undefined model: %s\n", model_name);
		}
		MPI_Bcast(&exit, 1, MPI_INT, 0, comm);
		if (exit) {
			MPI_Finalize();
			return 0;
		}
		MPIB_print_processors(comm);
		if (rank == 0) {
			CPM_print_model(model_name, msgset, parallel, precision, time);
			if (!strcmp(model_name, "Hockney"))
				Hockney_write(ostream, (Hockney_model*)model);
			if (!strcmp(model_name, "LinInterp"))
				LinInterp_write(ostream, (LinInterp_model*)model);
			if (!strcmp(model_name, "PLogP"))
				PLogP_write(ostream, (PLogP_model*)model);
#ifdef HAVE_LIBR
			if (!strcmp(model_name, "LMO"))
				LMO_write(ostream, (LMO_model*)model);
#endif /*HAVE_LIBR*/
		}
	} else {
		if (rank == 0) {
			if (!strcmp(model_name, "Hockney"))
				Hockney_read(istream, (Hockney_model**)&model);
			if (!strcmp(model_name, "LinInterp"))
				LinInterp_read(istream, (LinInterp_model**)&model);
			if (!strcmp(model_name, "PLogP"))
				PLogP_read(istream, (PLogP_model**)&model);
#ifdef HAVE_LIBR
			if (!strcmp(model_name, "LMO"))
				LMO_read(istream, (LMO_model**)&model);
#endif /*HAVE_LIBR*/
			exit = model == NULL;
		}
		MPIB_getopt_help_bcast(&exit, 0, comm);
		if (exit) {
			MPI_Finalize();
			return 0;
		}
	}

	if (rank == 0) {
		if (!strcmp(model_name, "Hockney")) {
			Hockney_model* hockney = (Hockney_model*)model;
			printf("#Homogeneous Hockney\n");
			printf("#a\t\tb\n");
			printf("#%le\t%le\n",
				gsl_stats_mean(hockney->a, 1, MPIB_C2(hockney->n)),
				gsl_stats_mean(hockney->b, 1, MPIB_C2(hockney->n)));
			printf("#Hockney prediction\n#msg\t");
			int i;
			for(i = 0; i < size - 1; i++) {
				int j;
				for(j = i + 1; j < size; j++)
					printf("%d-%d\t\t", i, j);
			}
			printf("\n");
			printf("0");
			for(i = 0; i < size - 1; i++) {
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", Hockney_predict_p2p(hockney, i, j, 0) * 2);
			}
			printf("\n");
			printf("%d", msgset.max_size);
			for(i = 0; i < size - 1; i++) {
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", Hockney_predict_p2p(hockney, i, j, msgset.max_size) * 2);
			}
			printf("\n\n\n");
		}
		/*
		  ToDo (optional): plot data for LinInterp here
		*/
		if (!strcmp(model_name, "PLogP")) {
			PLogP_model* plogp = (PLogP_model*)model; 
			printf("#PLogP prediction\n#msg\t");
			int i;
			for(i = 0; i < size - 1; i++) {
				int j;
				for(j = i + 1; j < size; j++)
					printf("%d-%d\t\t", i, j);
			}
			printf("\n");
			int M;
			for (M = 0; M < msgset.max_size; M += msgset.stride) {
				printf("%d", M);
				for(i = 0; i < size - 1; i++) {
					int j;
					for(j = i + 1; j < size; j++)
						printf("\t%le", PLogP_predict_p2p(plogp, i, j, M) * 2);
				}
				printf("\n");
			}
			printf("\n\n");
			printf("#LogGP prediction\n#msg\t");
			for(i = 0; i < size - 1; i++) {
				int j;
				for(j = i + 1; j < size; j++)
					printf("%d-%d\t\t", i, j);
			}
			printf("\n");
			printf("0");
			for(i = 0; i < size - 1; i++) {
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", LogGP_predict_p2p(&plogp->LogGP_predictor, i, j, 0) * 2);
			}
			printf("\n");
			printf("%d", msgset.max_size);
			for(i = 0; i < size - 1; i++) {
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", LogGP_predict_p2p(&plogp->LogGP_predictor, i, j, msgset.max_size) * 2);
			}
			printf("\n\n\n");
		}
#ifdef HAVE_LIBR
		if (!strcmp(model_name, "LMO")) {
			LMO_model* lmo = (LMO_model*)model;
			printf("#LMO prediction\n#msg\t");
			int i;
			for(i = 0; i < size - 1; i++) {
				int j;
				for(j = i + 1; j < size; j++)
					printf("%d-%d\t\t", i, j);
			}
			printf("\n");
			printf("0");
			for(i = 0; i < size - 1; i++) {
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", LMO_predict_p2p(lmo, i, j, 0) * 2);
			}
			printf("\n");
			printf("%d", msgset.max_size);
			for(i = 0; i < size - 1; i++) {
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", LMO_predict_p2p(lmo, i, j, msgset.max_size) * 2);
			}
			printf("\n\n\n");
		}
#endif /*HAVE_LIBR*/
	}

	if (!strcmp(model_name, "Hockney"))
		Hockney_free((Hockney_model*)model);
	if (!strcmp(model_name, "LinInterp"))
		LinInterp_free((LinInterp_model*)model);
	if (!strcmp(model_name, "PLogP"))
		PLogP_free((PLogP_model*)model);
#ifdef HAVE_LIBR
	if (!strcmp(model_name, "LMO"))
		LMO_free((LMO_model*)model);
#endif /*HAVE_LIBR*/

	if (rank == 0 && istream != NULL)
		fclose(istream);
	if (rank == 0 && ostream != stdout)
		fclose(ostream);

#ifndef MPIB_DEBUG
	MPI_Comm_free(&comm);
#endif
	MPI_Finalize();
	return 0;
}
